import os
from collections.abc import Callable, Coroutine
from functools import wraps
from typing import Any, ParamSpec, TypeVar

from opentelemetry import trace
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, ConsoleSpanExporter

from .settings import settings

resource = Resource(attributes={'service.name': settings.app_name})

trace.set_tracer_provider(TracerProvider(resource=resource))

devnull = open(os.devnull, 'w')
processor = BatchSpanProcessor(ConsoleSpanExporter(out=devnull))

if settings.opentelemetry_enable:  # pragma: no cover
    processor = BatchSpanProcessor(
        OTLPSpanExporter(
            endpoint=f'{settings.opentelemetry_scheme.value}://'
            f'{settings.opentelemetry_host}:{settings.opentelemetry_port}',
        ),
    )

tracer_provider: TracerProvider = trace.get_tracer_provider()  # type: ignore
tracer_provider.add_span_processor(processor)

P = ParamSpec('P')
T = TypeVar('T')


def decorator_trace(name: str = 'n/a'):
    """
    Decorator for the function, supports asynchronous functions,
    allows you to create a span inside the parent
        Example:
        .. code-block:: python
            @decorator_trace()
    """

    def decorator(function: Callable[P, Coroutine[Any, Any, T]]) -> Callable[P, Coroutine[Any, Any, T]]:
        @wraps(function)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            with trace.get_tracer(settings.app_name).start_as_current_span(name):
                return await function(*args, **kwargs)

        return wrapper

    return decorator



import typing

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.trustedhost import TrustedHostMiddleware

from . import handlers
from .logger import logger_init
from .middlewares import Middleware


class HelpTaskerCommonFastApiInstrumentator:
    def __init__(
        self,
        logger_load: bool = True,
        cors_enable: bool = False,
        cors_allow_origins: typing.Sequence[str] = (),
        cors_allow_credentials: bool = False,
        cors_allow_methods: typing.Sequence[str] = ('GET',),
        cors_allow_headers: typing.Sequence[str] = (),
        cors_max_age: int = 600,
        trusted_host_enable: bool = True,
        trusted_host_allowed_hosts: typing.Sequence[str] | None = (),
        healthcheck_enable: bool = False,
    ):
        self.logger_load = logger_load
        self.cors_enable = cors_enable
        self.cors_allow_origins = cors_allow_origins
        self.cors_allow_credentials = cors_allow_credentials
        self.cors_allow_methods = cors_allow_methods
        self.cors_allow_headers = cors_allow_headers
        self.cors_max_age = cors_max_age
        self.trusted_host_enable = trusted_host_enable
        self.trusted_host_allowed_hosts = trusted_host_allowed_hosts
        self.healthcheck_enable = healthcheck_enable

    def instrument(self, app: FastAPI):
        if self.logger_load:
            logger_init()

        app.add_middleware(Middleware)

        if self.cors_enable:
            app.add_middleware(
                CORSMiddleware,
                allow_origins=self.cors_allow_origins,
                allow_credentials=self.cors_allow_credentials,
                allow_methods=self.cors_allow_methods,
                allow_headers=self.cors_allow_headers,
                max_age=self.cors_max_age,
            )

        if self.trusted_host_enable:
            app.add_middleware(
                TrustedHostMiddleware,
                allowed_hosts=self.trusted_host_allowed_hosts,
            )

        if self.healthcheck_enable:
            app.include_router(handlers.healthcheck.router)

        return self



import time
import traceback
from uuid import uuid4

from fastapi import Request, status
from fastapi.responses import ORJSONResponse
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import Status, StatusCode
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint

from .context import cxt_api_version, cxt_ip
from .logger import cxt_request_id, logger
from .settings import Environment, settings
from .trace import trace


class Middleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint):
        api_version_token = cxt_api_version.set(0)
        cxt_ip_token = cxt_ip.set('0.0.0.0')
        cxt_request_id_token = cxt_request_id.set(request.headers.get('request-id', str(uuid4())))

        # Timer
        start_timer = time.monotonic()

        # IP
        if ip := request.headers.get('x-real-ip', request.headers.get('x-forwarded-for')):
            cxt_ip_token = cxt_ip.set(ip)

        # API Version
        if 'v1' in request.url.path.split('/', 3):
            api_version_token = cxt_api_version.set(1)
        elif 'v2' in request.url.path.split('/', 3):
            api_version_token = cxt_api_version.set(2)

        path = request.scope.get('path')
        method = request.scope.get('method')
        with trace.get_tracer(settings.app_name).start_as_current_span(f'{method} {path}') as span:
            try:
                response = await call_next(request)
            except Exception as err:
                if settings.environment == Environment.development:  # pragma: no cover
                    traceback.print_exc()

                response = ORJSONResponse(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                    content={'error': 'Internal server error'},
                )

                span_status_code = Status(status_code=StatusCode.ERROR, description=f'{err}')
                span.record_exception(exception=err)
            else:
                span_status_code = Status(status_code=StatusCode.OK)

            # Setting the status code
            span.set_status(span_status_code)

            # Setting attributes for trace
            span.set_attributes(
                {
                    SpanAttributes.HTTP_SCHEME: request.scope.get('scheme', 'n/a'),
                    SpanAttributes.HTTP_URL: request.url.path,
                    SpanAttributes.HTTP_CLIENT_IP: cxt_ip.get() or '0.0.0.0',
                    SpanAttributes.HTTP_USER_AGENT: request.headers.get('user-agent', 'n/a'),
                    SpanAttributes.HTTP_METHOD: request.method,
                    SpanAttributes.HTTP_STATUS_CODE: response.status_code,
                    SpanAttributes.HTTP_RESPONSE_CONTENT_LENGTH: int(response.headers.get('content-length', 0)),
                    'api.version': cxt_api_version.get(),
                    'request_id': cxt_request_id.get() or '',
                },
            )

            # Adding the header Request-ID server response
            response.headers['Request-ID'] = cxt_request_id.get() or ''

            # Adding the header Trace-ID server response
            if not span.get_span_context().trace_id == 0:
                response.headers['Trace-ID'] = format(span.get_span_context().trace_id, '032x')

            if request.url.path not in ['/metrics', '/healthcheck/']:
                extra = {
                    'method': request.method,
                    'user-agent': request.headers.get('user-agent', 'n/a'),
                    'ip': request.headers.get('x-real-ip', request.headers.get('x-forwarded-for', '127.0.0.1')),
                    'uri': request.url.path,
                    'response_status_code': response.status_code,
                    'request_duration_seconds': round(time.monotonic() - start_timer, 3),
                }

                if query := request.url.query:
                    extra['query'] = query

                logger.info(msg=f'{request.method} {request.url.path}', extra=extra)

        # Reset context var
        cxt_ip.reset(cxt_ip_token)
        cxt_api_version.reset(api_version_token)
        cxt_request_id.reset(cxt_request_id_token)
        return response



from contextlib import asynccontextmanager
from typing import AsyncGenerator, Optional

import httpx
from fastapi.applications import FastAPI
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import get_current_span

from . import __version__
from .logger import cxt_request_id, logger
from .settings import settings
from .trace import tracer_provider


@asynccontextmanager
async def http_base_client(
    base_url: str = '',
    verify: bool = True,
    accept: str = 'application/json',
    app: FastAPI | None = None,
    timeout: float | None = None,
    headers: dict[str, str] | None = None,
) -> AsyncGenerator[httpx.AsyncClient, None]:
    if headers is None:
        headers = {}

    headers['User-Agent'] = f'{settings.app_name}/{__version__}'
    headers['Accept-Language'] = 'en-US,en;q=0.9'
    headers['Accept'] = accept

    if request_id := cxt_request_id.get():
        headers['Request-ID'] = request_id

    client = httpx.AsyncClient(
        limits=httpx.Limits(
            max_keepalive_connections=settings.http_client_max_keepalive_connections,
            max_connections=settings.http_client_max_connections,
        ),
        base_url=base_url,
        http2=True,
        verify=verify,
        headers=headers,
        event_hooks={
            'request': [hook_request],
            'response': [hook_response],
        },
        timeout=timeout or settings.http_client_timeout,
        app=app,
    )

    try:
        with tracer_provider.get_tracer(settings.app_name).start_as_current_span('HTTPX HTTP request') as span:
            span.set_attribute('verify_certificate', verify)
            yield client
    finally:
        await client.aclose()


async def hook_request(request: httpx.Request):
    span = get_current_span()
    span.update_name(f'HTTPX - {request.method} {str(request.url.host)}')

    span.set_attributes(
        attributes={
            SpanAttributes.HTTP_URL: str(request.url),
            SpanAttributes.HTTP_METHOD: request.method,
            'request_id': request.headers.get('Request-ID', 0),
        },
    )  # type: ignore

    logger.info(f'Request http client: {request.method} {request.url} - Waiting for response')


async def hook_response(response: httpx.Response):
    span = get_current_span()
    span.set_attributes(
        attributes={
            SpanAttributes.HTTP_STATUS_CODE: response.status_code,
            SpanAttributes.HTTP_RESPONSE_CONTENT_LENGTH: response.headers.get('content-length', 0),
        },
    )

    request = response.request

    headers = {}
    for header in response.headers:
        headers[header] = response.headers.get(header)
    span.add_event('headers', attributes=headers)

    logger.info(f'Response http client: {request.method} {request.url} - Status {response.status_code}')



import os
from enum import Enum
from functools import cached_property
from pathlib import Path
from typing import Literal

from pydantic import PostgresDsn, SecretStr, computed_field
from pydantic_settings import BaseSettings, SettingsConfigDict


class LogLevel(str, Enum):
    """A class for determining logging levels"""

    debug: str = 'DEBUG'
    info: str = 'INFO'
    warning: str = 'WARNING'
    fatal: str = 'FATAL'


class Environment(str, Enum):
    """A class for determining environments"""

    development: str = 'DEVELOPMENT'
    testing: str = 'TESTING'
    production: str = 'PRODUCTION'


class Scheme(str, Enum):
    http: str = 'http'
    https: str = 'https'


class Settings(BaseSettings):
    app_name: str = 'FastID - Common'
    """ Application name """

    log_name: str = app_name
    """Logger name"""

    log_level: LogLevel = LogLevel.info
    """ Logging level """

    environment: Environment = Environment.production
    """ Environment """

    # Http client
    http_client_timeout: int = 10
    """HTTP defaults to including reasonable timeouts for all network operations, while Requests has no timeouts by
    default."""

    http_client_max_keepalive_connections: int = 5
    """ Number of allowable keep-alive connections """

    http_client_max_connections: int = 10
    """ Maximum number of allowable connections """

    # Tempo server
    opentelemetry_enable: bool = False
    """ Enables or disables sending """

    opentelemetry_scheme: Scheme = Scheme.http
    """Tempo scheme http or https"""

    opentelemetry_host: str = 'tempo'
    """Tempo host address as one of the following: an IP address or a domain name"""

    opentelemetry_port: int = 4317
    """Port number to connect to at the server host"""


settings = Settings()



import logging
import sys
from contextvars import ContextVar

from opentelemetry.trace import get_current_span
from pythonjsonlogger import jsonlogger

from .settings import settings

cxt_request_id: ContextVar[str | None] = ContextVar('request_id', default=None)

logger = logging.getLogger(settings.log_name)


class CustomJsonFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super().add_fields(log_record, record, message_dict)

        log_record['level'] = record.levelname
        log_record['logger'] = record.name

        if request_id := cxt_request_id.get():
            log_record['request_id'] = request_id

        current_span = get_current_span()
        if not current_span.get_span_context().trace_id == 0:
            log_record['span_id'] = format(current_span.get_span_context().span_id, '016x')
            log_record['trace_id'] = format(current_span.get_span_context().trace_id, '032x')


class InfoFilter(logging.Filter):
    def filter(self, rec):  # noqa: A003
        return rec.levelno in (logging.DEBUG, logging.INFO)


def logger_init() -> logging.Logger:
    formatter = CustomJsonFormatter()

    # StreamHandler
    handler_stream_stdout = logging.StreamHandler(sys.stdout)
    handler_stream_stderr = logging.StreamHandler(sys.stderr)

    handler_stream_stdout.setLevel(logging.DEBUG)
    handler_stream_stderr.setLevel(logging.WARNING)

    handler_stream_stdout.setFormatter(formatter)
    handler_stream_stderr.setFormatter(formatter)

    handler_stream_stdout.addFilter(InfoFilter())

    log = logging.getLogger(settings.log_name)
    log.setLevel(settings.log_level.upper())
    log.addHandler(handler_stream_stdout)
    log.addHandler(handler_stream_stderr)

    return log
